"""
Widget that holds the game tiles.
"""
import pygame.locals as pgl

from naja.constants import BOARD_SIZE, TILE_SIZE, KEYS, ACT, FPS
from naja.events import InvalidateTheWorld, finish_event
from naja.sound import sound

from naja.widgets.base import Widget
from naja.widgets.tile import TileWidget
from naja.utils import Flashlight


class BoardWidget(Widget):
    """
    Widget which holds all the tiles that make up the gameboard.
    """
    def __init__(self, pos, state, info):
        super(BoardWidget, self).__init__(pos, BOARD_SIZE)
        self.info = info
        self.state = state
        self.card_pos = state.player.position
        self._tiles = []
        self.legal = False
        self.flash_light = Flashlight(FPS // 2)
        for y in range(0, 5):
            for x in range(0, 5):
                tile_pos = (pos[0] + x * TILE_SIZE[0],
                            pos[1] + y * TILE_SIZE[1])
                self._tiles.append(TileWidget(tile_pos, state, (x, y)))

    def prepare(self):
        for tile in self._tiles:
            tile.set_highlight(self.card_pos, self.flash_light.on)
            tile.prepare()
        self.size = BOARD_SIZE
        if self.state.gameboard.player_mode == ACT:
            self.card_pos = self.state.player.position

    def draw(self, surface):
        if self.flash_light.tick():
            self.prepare()
        for tile in self._tiles:
            tile.draw(surface)

    def update_card_pos(self, card_pos):
        self.card_pos = card_pos
        self.info.set_position(self.card_pos)
        self.legal = self.card_pos in self.state.player.legal_moves()
        sound.play_sound('zoop.ogg', volume=0.05)

    def change_pos(self, offset):
        tpos = (offset[0] + self.card_pos[0], offset[1] + self.card_pos[1])
        if tpos[0] < 0 or tpos[1] < 0 or tpos[0] > 4 or tpos[1] > 4:
            return False
        self.update_card_pos(tpos)
        return True

    def next_viable_pos(self):
        moves = self.state.player.legal_moves()
        try:
            idx = moves.index(self.card_pos)
            idx = (idx + 1) % len(moves)
        except ValueError:
            idx = 0
        self.update_card_pos(moves[idx])

    def force_skip_animation(self):
        for tile in self._tiles:
            tile.animation = 0

    def handle_event(self, ev):
        if InvalidateTheWorld.matches(ev):
            self.flash_light.reset()
        if self.state.gameboard.player_mode == ACT:
            return super(BoardWidget, self).handle_event(ev)
        if ev.type == pgl.KEYDOWN:
            if ev.key in KEYS.UP:
                if self.change_pos((0, -1)):
                    return finish_event()
            elif ev.key in KEYS.DOWN:
                if self.change_pos((0, +1)):
                    return finish_event()
            elif ev.key in KEYS.LEFT:
                if self.change_pos((-1, 0)):
                    return finish_event()
            elif ev.key in KEYS.RIGHT:
                if self.change_pos((+1, 0)):
                    return finish_event()
            elif ev.key in KEYS.SWITCH:
                self.next_viable_pos()
                return finish_event()
            elif ev.key in KEYS.SELECT:
                if self.state.player.set_position(self.card_pos):
                    self.state.gameboard.change_mode(ACT)
                    self.info.set_position(self.card_pos)
                    sound.play_sound('chirp.ogg', volume=0.5)
                    return finish_event()
        return super(BoardWidget, self).handle_event(ev)
