"""
Widget for the game board information area.
"""
import pygame
import pygame.locals as pgl

from naja.constants import (
    INFO_SIZE, ACT, KEYS, EXAMINE, PALETTE,
    ACTION_TEXT_OFFSET, INFO_LEFT_PADDING,
    INFO_RIGHT_PADDING, BIT_SIZE, BITS)
from naja.events import InvalidateTheWorld, finish_event
from naja.resources import resources
from naja.resources.mutators import EIGHT_BIT, blender
from naja.sound import sound
from naja.utils import bit_glyphs, Flashlight

from naja.widgets.base import Widget
from naja.widgets.tile import BIT_MAP
from naja.widgets.text import TextBoxWidget
from naja import constants


HINTS = {
    ACT: "Choose an action using {NORTH,SOUTH}."
         " Press {RETURN} to execute it.",
    EXAMINE: "Browse the tiles with {NORTH,SOUTH,EAST,WEST} keys.",
}

HINT_LEGAL_MOVE = "\nPress {RETURN} to move."

TITLES = {
    ACT: "Choose an Action",
    EXAMINE: "Select a Tile",
}


class InfoAreaWidget(Widget):
    """
    Widget for the game board information area.
    """
    def __init__(self, pos, state):
        super(InfoAreaWidget, self).__init__(pos, INFO_SIZE)
        self.state = state
        self.chosen = None
        self.card_position = state.player.position
        self.set_position(state.player.position)
        self.flash_light = Flashlight(constants.FPS // 2)

    def prepare(self):
        if self.state.gameboard.player_mode == ACT:
            self.set_position(self.state.player.position)
        self.surface = pygame.surface.Surface(INFO_SIZE)
        self.surface.fill((0, 0, 0))

        box_width = INFO_SIZE[0] - INFO_LEFT_PADDING - INFO_RIGHT_PADDING
        y_offset = 0
        pos = lambda: (INFO_LEFT_PADDING, y_offset)

        # Bits
        y_offset += 12
        bits_text = ''.join('1' if bit in self.card.bitwise_operand else '0'
                            for bit in reversed(range(8)))
        if self.card.bitwise_operand:
            bits_text = '%s %s' % (
                bits_text, bit_glyphs(self.card.bitwise_operand))
        card_bits = TextBoxWidget(
            (0, 0), bits_text, padding=4, centre=True,
            colour=PALETTE.WHITE, border=2,
            bg_colour=PALETTE.BLACK, border_colour=PALETTE.BLUE,
            box_width=box_width)
        card_bits.prepare()
        self.surface.blit(card_bits.surface, pos())
        y_offset += card_bits.surface.get_rect().height + 12

        # Actions
        for choice, action in enumerate(self.card.actions):
            y_offset = self.prepare_action(choice, action, y_offset, box_width)

        # Hint text
        hint_text = HINTS[self.state.gameboard.player_mode]
        if self.state.gameboard.player_mode == EXAMINE:
            if self.card_position in self.state.player.legal_moves():
                hint_text += HINT_LEGAL_MOVE

        hint = TextBoxWidget(
            (0, 0), hint_text, padding=4,
            colour=PALETTE.WHITE, border=2,
            bg_colour=PALETTE.BLACK, border_colour=PALETTE.BLUE,
            box_width=box_width)
        hint.prepare()
        y_offset = (INFO_SIZE[1] - hint.surface.get_rect().height
                    - BIT_SIZE[1] - 2)
        self.surface.blit(hint.surface, pos())

        # Game mode (puzzle, easy, standard, hard, etc ...)
        game_mode = TextBoxWidget(
            (0, 0), self.mode_hint(), padding=4, centre=True,
            colour=PALETTE.WHITE, border=2,
            bg_colour=PALETTE.BLACK, border_colour=PALETTE.BLUE,
            box_width=box_width)
        game_mode.prepare()

        y_offset = (INFO_SIZE[1] - BIT_SIZE[1] + 12)
        self.surface.blit(game_mode.surface, pos())

    def mode_hint(self):
        gameboard = self.state.gameboard
        if gameboard.puzzle:
            return "PUZZLE"
        return {
            0: "DEATH", 1: "NINTENDO HARD",
            2: "VERY HARD", 3: "HARD",
            4: "STANDARD", 5: "EASY",
        }.get(gameboard.max_health, "UNKNOWN")

    def prepare_action(self, choice, action, y_offset, box_width):
        x_offset = INFO_LEFT_PADDING
        glyphs_x_offset = 2
        glyphs_y_offset = y_offset
        y_offset += ACTION_TEXT_OFFSET
        action_viable = action.check_available(self.state.player)
        text_colour = PALETTE.BLACK if action_viable else PALETTE.GREY

        text = TextBoxWidget(
            (x_offset, y_offset), action.get_text(self.card),
            box_width=box_width,
            fontsize=28, colour=text_colour,
            border=2, border_colour=PALETTE.GREY)
        text.render(self.surface)

        border_colour = None
        if choice == self.chosen:
            if self.flash_light.on:
                border_colour = (PALETTE.GREEN if action_viable else
                                 PALETTE.ORANGE)
            else:
                border_colour = (PALETTE.DARK_GREEN if action_viable else
                                 PALETTE.DARK_RED)

        if border_colour:
            bottom = y_offset + text.surface.get_rect().height
            right = text.surface.get_rect().width + x_offset
            pygame.draw.lines(self.surface, border_colour, True,
                              [(x_offset, y_offset), (right, y_offset),
                               (right, bottom), (x_offset, bottom)], 4)

        required_keys = action.required_bits & frozenset([
            BITS.RED, BITS.GREEN, BITS.BLUE])
        if required_keys in BIT_MAP:
            img_name = BIT_MAP[required_keys].replace(
                '.png', '_small.png')
            img = resources.get_image(img_name,
                                      transforms=(EIGHT_BIT,))
            self.surface.blit(img, (glyphs_x_offset, glyphs_y_offset))
            glyphs_x_offset += img.get_width() + 4
        else:
            glyphs_x_offset = INFO_LEFT_PADDING

        if BITS.MSB in action.required_bits:
            msb = resources.get_image('board/msb_lock_decoration.png',
                                      transforms=(EIGHT_BIT,))
            msb_rect = msb.get_rect()
            self.surface.blit(
                msb, (glyphs_x_offset - msb_rect.width - 4, glyphs_y_offset)
            )

        for glyph in action.get_glyphs():
            img = resources.get_image(
                glyph, transforms=(EIGHT_BIT, blender(PALETTE.GREY)))
            self.surface.blit(img, (glyphs_x_offset, glyphs_y_offset - 4))
            glyphs_x_offset += img.get_width()
        if action.get_msb_glyph() is not None:
            img = resources.get_image(
                action.get_msb_glyph(),
                transforms=(EIGHT_BIT, blender(PALETTE.LIGHT_VIOLET)))
            self.surface.blit(img, (glyphs_x_offset, glyphs_y_offset - 4))

        return y_offset + text.surface.get_rect().height + 16

    def set_position(self, position):
        self.card_position = position
        self.card = self.state.board_locations[self.card_position]
        if self.state.gameboard.player_mode == ACT:
            if self.chosen is None:
                self.chosen = 0
        else:
            self.chosen = None

    def draw(self, surface):
        if self.flash_light.tick():
            self.prepare()
        surface.blit(self.surface, self.pos)

    def next_action(self, viable_only=False, step=1):
        num_actions = len(self.card.actions)
        if num_actions == 0:
            return
        player = self.state.player
        chosen = self.chosen
        for i in range(num_actions - 1):
            # loop through each action at most once.
            chosen = (chosen + step) % num_actions
            action = self.card.actions[chosen]
            if not viable_only or action.check_available(player):
                sound.play_sound('zoop.ogg', volume=0.05)
                self.chosen = chosen

    def prev_action(self, viable_only=False):
        return self.next_action(viable_only=viable_only, step=-1)

    def try_perform_action(self):
        player = self.state.player
        action = self.card.actions[self.chosen]
        if not action.check_available(player):
            sound.play_sound('error.ogg')
        else:
            sound.play_sound('chirp.ogg', volume=0.5)
            action.perform_action(self.state.gameboard, self.card)
            self.state.gameboard.card_used(player.position)
            self.state.gameboard.change_mode(EXAMINE)
            self.set_position(player.position)

    def handle_event(self, ev):
        if InvalidateTheWorld.matches(ev):
            self.flash_light.reset()
        if self.state.gameboard.player_mode == EXAMINE:
            return super(InfoAreaWidget, self).handle_event(ev)
        if ev.type == pgl.KEYDOWN:
            if ev.key in KEYS.SELECT:
                self.try_perform_action()
                return finish_event()
            if ev.key in KEYS.UP:
                self.next_action()
                return finish_event()
            if ev.key in KEYS.DOWN:
                self.prev_action()
                return finish_event()
            if ev.key in KEYS.SWITCH:
                self.next_action(viable_only=True)
                return finish_event()
        return super(InfoAreaWidget, self).handle_event(ev)
