import pygame
import pygame.locals as pgl

from naja.constants import FONT, FONT_SIZE, EIGHT_BIT_SCALE, PALETTE, KEYS
from naja.resources import resources
from naja.resources.mutators import EIGHT_BIT, R180, blender
from naja.utils import convert_colour
from naja.widgets.base import Widget


MARKUP_MAP = {
    'NORTH': ('glyphs/arrow_up.png', None),
    'SOUTH': ('glyphs/arrow_down.png', None),
    'EAST': ('glyphs/arrow_right.png', None),
    'WEST': ('glyphs/arrow_left.png', None),
    'RETURN': ('glyphs/return.png', None),
    'HEALTH': ('glyphs/health.png', PALETTE.DARK_RED),
    'WINTOKEN': ('glyphs/win.png', PALETTE.DARK_OLIVE),
    'KEY': ('glyphs/key.png', None),
    'MSB': ('glyphs/msb.png', None),
    'RED': ('glyphs/key.png', PALETTE.ORANGE),
    'GREEN': ('glyphs/key.png', PALETTE.GREEN),
    'BLUE': ('glyphs/key.png', PALETTE.BLUE),
    'CLOCKWISE': ('glyphs/clockwise.png', None),
    'ANTICLOCKWISE': ('glyphs/anticlockwise.png', None),
    'SHIFT_LEFT': ('glyphs/shift_left.png', None),
    'SHIFT_RIGHT': ('glyphs/shift_right.png', None),
    'COUNTDOWN': ('glyphs/countdown_4.png', PALETTE.DARK_VIOLET),

    'HEALTH_NOCOLOUR': ('glyphs/health.png', None),
    'WINTOKEN_NOCOLOUR': ('glyphs/win.png', None),
}


class Glyph(object):
    def __init__(self, markup_text, glyph_keys, suffix=''):
        self.markup_text = markup_text
        self.glyph_keys = glyph_keys
        self.text = ' ' * len(self.glyph_keys) + suffix


class TextWidget(Widget):

    VIEW_PORT_DY = 50

    def __init__(self, pos, text, fontname=None, fontsize=None,
                 colour=None, unselectable_colour=None, view_port=None,
                 centre=False, border=0, border_colour=PALETTE.GREY,
                 padding=2):
        super(TextWidget, self).__init__(pos)

        self.text = text
        self.fontname = fontname or FONT
        self.fontsize = (fontsize or FONT_SIZE) // EIGHT_BIT_SCALE
        self.colour = convert_colour(colour or PALETTE.BLACK)
        if unselectable_colour is not None:
            unselectable_colour = convert_colour(unselectable_colour)
        self.unselectable_colour = unselectable_colour
        self.view_port = (
            pygame.Rect((0, 0), view_port) if view_port is not None else None)
        self.centre = centre
        self.centre_pos = pos
        self.border = border
        self.border_colour = convert_colour(border_colour)
        self.padding = padding

    def render_line(self, text):
        colour = self.colour
        if not self.is_selectable() and self.unselectable_colour is not None:
            colour = self.unselectable_colour
        text_surf = self.font.render(text, True, colour)
        text_rect = text_surf.get_rect()
        return pygame.transform.scale(
            text_surf, (text_rect.width * EIGHT_BIT_SCALE,
                        text_rect.height * EIGHT_BIT_SCALE))

    def render_border(self, surface):
        if not self.border or not self.border_colour:
            return
        rect = surface.get_rect()
        rect = rect.inflate(- self.border / 2, - self.border / 2)
        pygame.draw.rect(surface, self.border_colour, rect, self.border)

    def update_size(self):
        if self.view_port is not None:
            self.size = self.view_port.size
        else:
            self.size = self.surface.get_rect().size
        if self.centre:
            self.pos = (self.centre_pos[0] - self.size[0] // 2,
                        self.centre_pos[1])

    def prepare(self):
        self.font = resources.get_font(self.fontname, self.fontsize)
        self.surface = self.render_line(self.text)
        self.render_border(self.surface)
        self.update_size()

    def handle_event(self, ev):
        if self.view_port is None:
            return super(TextWidget, self).handle_event(ev)
        if ev.type == pgl.KEYDOWN:
            if ev.key in KEYS.DOWN:
                self.view_port.move_ip(0, self.VIEW_PORT_DY)
                if self.view_port.bottom > self.surface.get_rect().bottom:
                    self.view_port.bottom = self.surface.get_rect().bottom
                return True
            elif ev.key in KEYS.UP:
                self.view_port.move_ip(0, -self.VIEW_PORT_DY)
                if self.view_port.top < 0:
                    self.view_port.top = 0
                return True
        return super(TextWidget, self).handle_event(ev)

    def draw(self, surface):
        if self.view_port is None:
            rect = self.rect
            area = None
        else:
            rect = self.pos
            area = self.view_port
        surface.blit(self.surface, rect, area)
        if self.view_port is not None:
            self.draw_arrows(surface)

    def draw_arrows(self, surface):
        if self.view_port.top > 0:
            up = resources.get_image('bits', 'arrow_on.png',
                                     transforms=(EIGHT_BIT,))
            icon_size = up.get_rect().height
            pos = (self.pos[0] + self.view_port.width - icon_size, self.pos[1])
            surface.blit(up, pos)
        if self.view_port.bottom < self.surface.get_rect().bottom:
            down = resources.get_image('bits', 'arrow_on.png',
                                       transforms=(R180, EIGHT_BIT))
            icon_size = down.get_rect().height
            pos = (self.pos[0] + self.view_port.width - icon_size,
                   self.pos[1] + self.view_port.height - icon_size)
            surface.blit(down, pos)


class TextBoxWidget(TextWidget):
    def __init__(self, *args, **kwargs):
        self.bg_colour = convert_colour(kwargs.pop('bg_colour',
                                                   PALETTE.LIGHT_VIOLET))
        self.box_width = kwargs.pop('box_width', 0)
        self.full_width = kwargs.pop('full_width', True)
        kwargs.setdefault('padding', 4)

        super(TextBoxWidget, self).__init__(*args, **kwargs)

    def lines(self):
        if self.box_width != 0:
            return self._wrapped_lines()
        else:
            return ((line, []) for line in self.text.splitlines())

    def _prepare_glyph(self, glyph, current_words):
        glyphs = []
        size = self.font.size(' '.join(current_words[:-1] + ['']))
        x = size[0] * EIGHT_BIT_SCALE
        for glyph_key in glyph.glyph_keys:
            image_name, colour = MARKUP_MAP[glyph_key]
            if colour is None:
                colour = self.colour
            image = resources.get_image(
                image_name, transforms=(EIGHT_BIT, blender(colour)))
            glyphs.append(((x, 0), image))
            x += image.get_width()
        return glyphs

    def _check_markup(self, word):
        suffix = ''
        if word[-1] in '.,:':
            suffix = word[-1]
            word = word[:-1]

        if not word:
            return None
        if word[0] == '{' and word[-1] == '}':
            subwords = word[1:-1].split(',')
            if all(subword in MARKUP_MAP for subword in subwords):
                return Glyph(word + suffix, subwords, suffix)

        return None

    def _wrapped_lines(self):
        def words_fit(words):
            words_line = ' '.join(words)
            width = self.font.size(words_line)[0] * EIGHT_BIT_SCALE
            if width < self.box_width:
                return True
            elif len(words) == 1:
                Exception("Word %r too long for box." % (words[0],))
            return False

        line_count = 0
        for line in self.text.splitlines():
            line = line.strip()
            if not line:
                line_count += 1
                yield (line, [])
                continue
            current_words, glyphs = [], []
            remaining_words = line.split()
            while remaining_words:
                word = remaining_words.pop(0)
                glyph = self._check_markup(word)
                if glyph is not None:
                    word = glyph.text
                current_words.append(word)
                if words_fit(current_words):
                    if glyph is not None:
                        glyphs.extend(self._prepare_glyph(
                            glyph, current_words))
                else:
                    line_count += 1
                    yield (' '.join(current_words[:-1]), glyphs)
                    current_words, glyphs = [], []
                    if glyph is not None:
                        word = glyph.markup_text
                    remaining_words.insert(0, word)
            if current_words and words_fit(current_words):
                line_count += 1
                yield (' '.join(current_words), glyphs)

    def prepare(self):
        self.font = resources.get_font(self.fontname, self.fontsize)
        rendered_lines = []
        width, height = self.padding * 2, self.padding * 2
        for line, glyphs in self.lines():
            line_surface = self.render_line(line)
            line_rect = line_surface.get_rect()
            rendered_lines.append(line_surface)
            width = max(width, line_rect.width + self.padding * 2)
            height += line_rect.height
            for pos, img in glyphs:
                line_surface.blit(img, pos)

        if self.full_width:
            width = max(width, self.box_width)

        self.surface = pygame.surface.Surface((width, height),
                                              pygame.locals.SRCALPHA)
        self.surface.fill(self.bg_colour)
        self.update_size()

        x, y = self.padding, self.padding
        for line_surface in rendered_lines:
            if self.centre:
                x = (width - line_surface.get_rect().width) / 2
                x += self.padding
            self.surface.blit(line_surface, (x, y))
            y += line_surface.get_rect().height

        self.render_border(self.surface)
